package de.jstacs.sequenceScores.statisticalModels.trainable.latentVariables;

import java.text.NumberFormat;
import java.util.Random;
import javax.naming.OperationNotSupportedException;
import de.jstacs.NotTrainedException;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.DataSet;
import de.jstacs.data.EmptyDataSetException;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.NonParsableException;
import de.jstacs.io.XMLParser;
import de.jstacs.motifDiscovery.MotifDiscoverer;
import de.jstacs.results.NumericalResultSet;
import de.jstacs.sequenceScores.statisticalModels.trainable.AbstractTrainableStatisticalModel;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.AbstractVariableStructureModel;
import de.jstacs.utils.Normalisation;
import de.jstacs.utils.algorithms.BasicSamplingAlgorithms;
import de.jstacs.utils.functions.NML;


/**
 * This class represents a ZOOPS model for representing DNA sequence that contain a recurring pattern (sequence motif).
 * @author Ralf Eggeling <eggeling@informatik.uni-halle.de>
 *
 */
public class ZOOPSModel extends AbstractLatentVariableModel implements MotifDiscoverer {
	
	private static final String XML_TAG = "ZOOPSModel";

	private AbstractVariableStructureModel motifModel;
	
	private AbstractTrainableStatisticalModel flankingModel;
	
	private double[] motifOccurrenceLogProbabilityParameter;
	
	private double[] strandOrientationLogProbabilityParameter;
	
	private boolean textOutput = true;
	
	/**
	 * Creates ZOOPS model for learning via stochastic maximization algorithm
	 * @param motifModel The model for the putative binding sites
	 * @param flankingModel The model for the flanking regions
	 * @param numberOfIterations Number of iteration steps per restart
	 * @param numberOfRestarts Number of restarts
	 * @throws Exception
	 */
	public ZOOPSModel(AbstractVariableStructureModel motifModel, AbstractTrainableStatisticalModel flankingModel, int numberOfInitialIterations,  int numberOfAdditionalIterations, int numberOfRestarts) throws Exception {
		super(new AlphabetContainer(DNAAlphabet.SINGLETON),0);
		this.motifModel = motifModel.clone();
		this.flankingModel = flankingModel.clone();
		this.numberOfInitialIterations = numberOfInitialIterations;
		this.numberOfAdditionalIterations = numberOfAdditionalIterations;
		this.numberOfRestarts = numberOfRestarts;
		motifOccurrenceLogProbabilityParameter = new double[]{Math.log(0),Math.log(1)};
		strandOrientationLogProbabilityParameter = new double[]{Math.log(0.5),Math.log(0.5)};
		this.length=0;
		mode = STOCHASTIC_EM;
	}
	
	/**
	 * Creates ZOOPS model for learning via Gibbs sampling algorithm for model averaging
	 * @param motifModel The model for the putative binding sites
	 * @param flankingModel The model for the flanking regions
	 * @param numberOfIterations Number of iteration steps per restart
	 * @param numberOfRestarts Number of restarts
	 * @param lengthOfBurnIn The number of iteration steps until parameter samples are stored
	 * @param lagInStationaryPhase The number of iteration steps between two stored parameter samples
	 * @throws Exception
	 */
	public ZOOPSModel(AbstractVariableStructureModel motifModel, AbstractTrainableStatisticalModel flankingModel, int numberOfIterations, int numberOfRestarts, int lengthOfBurnIn, int lagInStationaryPhase) throws Exception {
		super(new AlphabetContainer(DNAAlphabet.SINGLETON),0);
		this.motifModel = motifModel.clone();
		this.flankingModel = flankingModel.clone();
		this.numberOfInitialIterations = numberOfIterations;
		this.numberOfRestarts = numberOfRestarts;
		motifOccurrenceLogProbabilityParameter = new double[]{Math.log(0),Math.log(1)};
		strandOrientationLogProbabilityParameter = new double[]{Math.log(0.5),Math.log(0.5)};
		this.length=0;
		this.lengthOfBurnIn = lengthOfBurnIn;
		this.lagInStationaryPhase = lagInStationaryPhase; 
		mode = GIBBS_SAMPLING;
	}
	
	/**
	 * Loads a ZOOPS model from XML
	 * @param sb XML representation
	 * @throws NonParsableException
	 */
	public ZOOPSModel(StringBuffer sb) throws NonParsableException {
		super(XMLParser.extractForTag(sb,XML_TAG));
	}
	
	protected void fromXML(StringBuffer xml) throws NonParsableException {
		this.alphabets = XMLParser.extractObjectForTags(xml, "alphabets", AlphabetContainer.class);
		this.length = XMLParser.extractObjectForTags(xml, "length", int.class);
		this.flankingModel =  XMLParser.extractObjectForTags(xml, "flankingModel", AbstractTrainableStatisticalModel.class);
		this.motifModel = XMLParser.extractObjectForTags(xml, "motifModel", AbstractVariableStructureModel.class);
		this.strandOrientationLogProbabilityParameter = XMLParser.extractObjectForTags(xml, "strandParameter", double[].class);
		this.motifOccurrenceLogProbabilityParameter = XMLParser.extractObjectForTags(xml, "occurrenceParameter", double[].class);
		this.numberOfInitialIterations =  XMLParser.extractObjectForTags(xml, "numberOfInitialIterations", int.class);
		this.numberOfAdditionalIterations =  XMLParser.extractObjectForTags(xml, "numberOfAdditionalIterations", int.class);
		this.numberOfRestarts = XMLParser.extractObjectForTags(xml, "numberOfRestarts", int.class);
	}

	/**
	 * Writes an XML representation of the ZOOPS model
	 * @throws NonParsableException
	 */
	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, alphabets, "alphabets");
		XMLParser.appendObjectWithTags(xml, length, "length");
		XMLParser.appendObjectWithTags(xml, flankingModel, "flankingModel");
		XMLParser.appendObjectWithTags(xml, motifModel, "motifModel");
		XMLParser.appendObjectWithTags(xml, strandOrientationLogProbabilityParameter, "strandParameter");
		XMLParser.appendObjectWithTags(xml, motifOccurrenceLogProbabilityParameter, "occurrenceParameter");
		XMLParser.appendObjectWithTags(xml, numberOfInitialIterations, "numberOfInitialIterations");
		XMLParser.appendObjectWithTags(xml, numberOfAdditionalIterations, "numberOfAdditionalIterations");
		XMLParser.appendObjectWithTags(xml, numberOfRestarts, "numberOfRestarts");
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}
	
	public ZOOPSModel clone() throws CloneNotSupportedException {
		return (ZOOPSModel) super.clone();
	}
	
	public void train(DataSet data, double[] weights) throws Exception {
				
		int[] hiddenVariableMotifOccurrence = new int[data.getNumberOfElements()];
		int[] hiddenVariableBindingSitePosition =  new int[data.getNumberOfElements()];
		int[] hiddenVariableStrandOrientation = new int[data.getNumberOfElements()];
		int[] bestHVMO = new int[data.getNumberOfElements()];
		int[] bestHVBSP = new int[data.getNumberOfElements()];
		int[] bestHVSO = new int[data.getNumberOfElements()];
		if(weights!=null)
			System.out.println("Weights are ignored");
		Random rand = new Random();
		if(!flankingModel.isInitialized())
			flankingModel.train(data);
		if(mode==GIBBS_SAMPLING) {
			motifModel.initForSampling(sampleStorageDirectory+"/"+expString+"_motifParams.txt" ) ;
		}
		double maxScore =Double.NEGATIVE_INFINITY;
		int argmaxIt = 0;
		int argmaxRs = 0;
		for(int rs = 0;rs<numberOfRestarts;rs++) {
			//initialization
			double maxScoreLocal = Double.NEGATIVE_INFINITY;
			for(int i=0;i<data.getNumberOfElements();i++) {
				hiddenVariableMotifOccurrence[i] = BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromLogProbs(motifOccurrenceLogProbabilityParameter, rand);
		//		hiddenVariableBindingSitePosition[i] = BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromUniformDistribution(data.getElementAt(i).getLength()-motifModel.getLength()+1, rand);
				hiddenVariableBindingSitePosition[i] = rand.nextInt(data.getElementAt(i).getLength()-motifModel.getLength()+1);
				hiddenVariableStrandOrientation[i] =  BasicSamplingAlgorithms.sampleDiscreteRandomVariableFromLogProbs(strandOrientationLogProbabilityParameter, rand);
			}
			
			DataSet currentBindingSites; 
			double previousScore = 0;
			int lastOptimalModelUpdate = 0;
			for(int it = 0; it<numberOfInitialIterations || mode==STOCHASTIC_EM&&((it-lastOptimalModelUpdate)<numberOfAdditionalIterations);it++) {
				if(textOutput)
					System.out.print(rs+"\t"+it+"\t");
				// compute or sample parameters
				currentBindingSites = this.getBindingSitesForLatentVariables(data, hiddenVariableMotifOccurrence,  hiddenVariableBindingSitePosition,hiddenVariableStrandOrientation);
				if(mode==STOCHASTIC_EM) {
					motifModel.train(currentBindingSites);
					this.estimateHVParameters(hiddenVariableMotifOccurrence, hiddenVariableStrandOrientation);		
				} 
				if(mode==GIBBS_SAMPLING) {
					motifModel.drawParameters(currentBindingSites);
					if(it>=this.lengthOfBurnIn) {
						motifModel.saveCurrentParametersToFile(); 
					}
				}
				// sample latent variables
				double LL = 0;
				int[] lv = new int[data.getNumberOfElements()];
				for(int i=0;i<data.getNumberOfElements();i++) {
					double[] profile = getUnnormalizedSequenceProfile(data.getElementAt(i));
					LL += Normalisation.logSumNormalisation(profile);
					lv[i] = BasicSamplingAlgorithms.sampleDiscreteRandomVariable(profile, rand);
				}	
					double score = motifModel.getLogScoreOfCurrentStructure();
					if(score>maxScoreLocal) {
						maxScoreLocal=score;
						lastOptimalModelUpdate = it;
						if(score>maxScore) {
							maxScore=score;
							argmaxIt = it;
							argmaxRs = rs;
							bestHVMO = hiddenVariableMotifOccurrence.clone();
							bestHVBSP = hiddenVariableBindingSitePosition.clone();
							bestHVSO = hiddenVariableStrandOrientation.clone();
						}
					}
					if(textOutput) {
						if(mode==STOCHASTIC_EM)
							System.out.print(score+"\t"+(score-previousScore)+"\t"+argmaxRs+"\t"+argmaxIt+"\t"+LL+"\t"+motifModel.getNumericalCharacteristics().getResultAt(0).getValue());
						if(mode==GIBBS_SAMPLING)
							System.out.print(LL+"\t"+motifModel.getNumericalCharacteristics().getResultAt(0).getValue());
					}
						
					previousScore=score;
					for(int i=0;i<data.getNumberOfElements();i++) {
						if(lv[i]==0) {
							hiddenVariableMotifOccurrence[i] = 0;
						} else {
							hiddenVariableMotifOccurrence[i] = 1;
							hiddenVariableBindingSitePosition[i] = (int)Math.floor((lv[i]-1)/2);
							hiddenVariableStrandOrientation[i] = 1-((2*hiddenVariableBindingSitePosition[i]-lv[i])+2);
						}
					}
					if(textOutput)
						System.out.println();
			}

		}
		
		if(mode==STOCHASTIC_EM) {
			hiddenVariableMotifOccurrence = bestHVMO.clone();
			hiddenVariableBindingSitePosition=bestHVBSP.clone();
			hiddenVariableStrandOrientation=bestHVSO.clone();
			DataSet currentBindingSites = this.getBindingSitesForLatentVariables(data, hiddenVariableMotifOccurrence,  hiddenVariableBindingSitePosition,hiddenVariableStrandOrientation);
			motifModel.train(currentBindingSites);
			this.estimateHVParameters(hiddenVariableMotifOccurrence, hiddenVariableStrandOrientation);
			if(textOutput) {
				System.out.println("OptimalModel:");
				System.out.println("Restart="+argmaxRs);
				System.out.println("Iteration step="+argmaxIt);
				System.out.println("Structure score="+motifModel.getLogScoreOfCurrentStructure());
				System.out.println("Likelihood="+this.getLogProbFor(data));
			}
					
		} 
		if(mode==GIBBS_SAMPLING) {
			motifModel.samplingStopped() ;
		}
				
	}
	
	
	/**
	 * Returns the set of binding sites that is associated with a given set of hidden variables.
	 * @param data
	 * @param hvmo
	 * @param hvbsp
	 * @param hvso
	 * @return
	 * @throws EmptyDataSetException
	 * @throws WrongAlphabetException
	 */
	private DataSet getBindingSitesForLatentVariables(DataSet data, int[] hvmo, int[] hvbsp, int[] hvso) throws EmptyDataSetException, WrongAlphabetException {
		int numberOfBS=0;
		for(int i=0;i<hvmo.length;i++) {
			if(hvmo[i]==1)
				numberOfBS++;
		}
		Sequence[] newBS = new Sequence[numberOfBS];
		int n = 0;
		for(int i=0;i<data.getNumberOfElements();i++) {
			if(hvmo[i]==1) {
				if(hvso[i]==0) {
					newBS[n] = data.getElementAt(i).getSubSequence(hvbsp[i], motifModel.getLength());
				}				
				if(hvso[i]==1) {
					try {
						newBS[n] = data.getElementAt(i).getSubSequence(hvbsp[i], motifModel.getLength()).reverseComplement();
					} catch(OperationNotSupportedException e) {
						System.out.println("Warning: Reverse complement not supported. Use forward strand instead.");
						newBS[n] = data.getElementAt(i).getSubSequence(hvbsp[i], motifModel.getLength());
					}
					
				}
				n++;
			}
		}
		return new DataSet("",newBS);
	}
	
	/**
	 * Estimates parameters for motif occurrence and strand probability from the current state of hidden variables
	 * @param hiddenVariableMotifOccurrence Hidden variables indicating motif occurrence per sequence
	 * @param hiddenVariableStrandOrientation Hidden variables indiciating strand orientation per sequence 
	 */
	private void estimateHVParameters(int[] hiddenVariableMotifOccurrence, int[] hiddenVariableStrandOrientation) {
		int forward = 0;
		int backward = 0;
		int site = 0;
		int noSite = 0;
		for(int i=0;i<hiddenVariableMotifOccurrence.length;i++) {
			if(hiddenVariableMotifOccurrence[i]==1) {
				site++;
				if(hiddenVariableStrandOrientation[i] == 0) {
					forward++;

				} else {
					backward++;
				}
			} else {
				noSite++;
			}
		}
		motifOccurrenceLogProbabilityParameter = new double[]{NML.eTerm(noSite)*(noSite+1),NML.eTerm(site)*(site+1)};
		Normalisation.sumNormalisation(motifOccurrenceLogProbabilityParameter);
		motifOccurrenceLogProbabilityParameter[0] = Math.log(motifOccurrenceLogProbabilityParameter[0]);
		motifOccurrenceLogProbabilityParameter[1] = Math.log(motifOccurrenceLogProbabilityParameter[1]);
		strandOrientationLogProbabilityParameter = new double[]{NML.eTerm(forward)*(forward+1),NML.eTerm(backward)*(backward+1)};
		Normalisation.sumNormalisation(strandOrientationLogProbabilityParameter);
		strandOrientationLogProbabilityParameter[0] = Math.log(strandOrientationLogProbabilityParameter[0]);
		strandOrientationLogProbabilityParameter[1] = Math.log(strandOrientationLogProbabilityParameter[1]);
	}
	
	
	public double getLogPriorTerm() throws Exception {
		return 0;
	}
	
	public double getLogProbFor(Sequence sequence, int startpos, int endpos) throws Exception {
	//	if(mode==0) {
			return Normalisation.logSumNormalisation(getUnnormalizedSequenceProfile(sequence.getSubSequence(startpos, endpos-startpos+1)));
	/*	} else {
			reader = new BufferedReader( new FileReader( new File(sampleStorageDirectory+"/motifParams.txt") ) );
			int N = 0;
			while(reader.ready()) {
				N++;
				reader.readLine();
			}
			reader =  new BufferedReader( new FileReader( new File(sampleStorageDirectory+"/motifParams.txt") ) );
			double[] res = new double[N];
			for(int i = 0;i<N;i++) {
				motifModel.parse(reader.readLine());
				
				res[i] += Normalisation.logSumNormalisation(getUnnormalizedSequenceProfile(sequence.getSubSequence(startpos, endpos-startpos+1)));
		//		System.out.println( Normalisation.logSumNormalisation(getUnnormalizedSequenceProfile(sequence.getSubSequence(startpos, endpos-startpos+1))));
			}
			return Normalisation.logSumNormalisation(res) - Math.log(N);
			
		}*/
		
	}
	
	public double getLogProbFor(DataSet data) throws Exception {
		double res = 0;
		for(int i=0;i<data.getNumberOfElements();i++) {
			res+=this.getLogProbFor(data.getElementAt(i),0,data.getElementAt(i).getLength()-1);
		}
		return res;
	}

	public String getInstanceName() {
		return "ZOOPS";
	}

	public NumericalResultSet getNumericalCharacteristics() throws Exception {
		return null;
	}

	public boolean isInitialized() {
		return flankingModel.isInitialized()&&motifModel.isInitialized();
		//return trained;
	}

	public int getMotifLength(int motif) {
		return motifModel.getLength();
	}

	public int getNumberOfComponents() {
		return 1;
	}

	public int getNumberOfMotifs() {
		return 1;
	}

	public int getNumberOfMotifsInComponent(int component) {
			switch( component ) {
			case 0:
				return 1; //component with motif
			case 1:
				return 0; //component without component
			default:
				throw new IndexOutOfBoundsException();
			}
	}


	public int getGlobalIndexOfMotifInComponent(int component, int motif) {
		if( component == 0 && motif == 0 ) {
			return 0;
		} else {
			throw new IndexOutOfBoundsException();
		}
	}
	
	public double[] getProfileOfScoresFor(int component, int motif,	Sequence sequence, int startpos, KindOfProfile kind) throws Exception {
	//	System.out.println(kind.toString());
		if( component == 0 && motif == 0 ) {
			Sequence seq = sequence.getSubSequence(startpos);
			double[] res = new double[sequence.getLength()-startpos-motifModel.getLength()+1];
			for(int i=0;i<res.length;i++) {
	//			res[i]+=flankingModel.getLogProbFor(seq, 0, i-1);
	//			res[i]+=flankingModel.getLogProbFor(seq, i+motifModel.getLength(), seq.getLength()-1);
				
				double[] bsLogProb = new double[]{motifModel.getLogProbFor(seq.getSubSequence(i,motifModel.getLength()))+strandOrientationLogProbabilityParameter[0],
				motifModel.getLogProbFor(seq.getSubSequence(i,motifModel.getLength()).reverseComplement())+strandOrientationLogProbabilityParameter[1]};
				res[i]+=Normalisation.getLogSum(bsLogProb);
		//		System.out.print(res[i]+"\t");
			}
		//	System.out.println();
		//	if( kind == KindOfProfile.NORMALIZED_CONDITIONAL ) {
	//			double d = Normalisation.getLogSum( res );
	//			for( int i=0;i<res.length;i++ ) {
	//				res[i] -= d;
		//			System.out.print(res[i]+"\t");
	//			}
		//		System.out.println();
		//	}
			
/*			if( kind == KindOfProfile.NORMALIZED_CONDITIONAL ) {
				double norm = Normalisation.getLogSum(res);
				for(int i=0;i<res.length;i++) {
					res[i] -=norm;
				}
			} */
			return res;
			} else {
			throw new IndexOutOfBoundsException();
		}

	}
	
	/** 
	 * Returns the log probability for each possible start position (both strands). 0 is no motif occurrence. odd numbers encode forward strand positions, even numbers encode reverse strand positions.
	 * @param sequence
	 * @param startpos
	 * @param length
	 * @throws Exception 
	 * @throws NotTrainedException 
	 * 
	 */
	private double[] getUnnormalizedSequenceProfile(Sequence sequence) throws NotTrainedException, Exception {
		double[] res = new double[2*(sequence.getLength()-motifModel.getLength()+1)+1];
		res[0] = motifOccurrenceLogProbabilityParameter[0] + flankingModel.getLogProbFor(sequence);
		for(int i=0;i<sequence.getLength()-motifModel.getLength()+1;i++) {
	//		System.out.println((i+motifModel.getLength())+"\t"+(sequence.getLength()-1));
			res[2*i+1] = motifOccurrenceLogProbabilityParameter[1]
					   + strandOrientationLogProbabilityParameter[0]
					   + flankingModel.getLogProbFor(sequence, 0, i-1)
			           + motifModel.getLogProbFor(sequence.getSubSequence(i,motifModel.getLength()))
			           + flankingModel.getLogProbFor(sequence, i+motifModel.getLength(), sequence.getLength()-1);
			res[2*i+2] = motifOccurrenceLogProbabilityParameter[1]
					   + strandOrientationLogProbabilityParameter[1]
					   + flankingModel.getLogProbFor(sequence, 0, i-1)
					   + motifModel.getLogProbFor(sequence.getSubSequence(i,motifModel.getLength()).reverseComplement())
		               + flankingModel.getLogProbFor(sequence, i+motifModel.getLength(), sequence.getLength()-1);
			}
		return res;
	}
	
	public double[] getStrandProbabilitiesFor(int component, int motif,	Sequence sequence, int startpos) throws Exception {
		 if( component == 0 && motif == 0 ) {
			double[] res = new double[]{
			strandOrientationLogProbabilityParameter[0]+motifModel.getLogProbFor(sequence.getSubSequence(startpos,motifModel.getLength())),strandOrientationLogProbabilityParameter[1]+motifModel.getLogProbFor(sequence.getSubSequence(startpos,motifModel.getLength()).reverseComplement())};
			Normalisation.logSumNormalisation(res);
			return res;
	        } else {
	            throw new IndexOutOfBoundsException();
	        }
	}
	
	public AbstractTrainableStatisticalModel getMotifModel() {
		return motifModel;
	}
	
	
	public AbstractTrainableStatisticalModel getFlankingModel() {
		return flankingModel;
	}

	@Override
	public String toString(NumberFormat nf) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public int getIndexOfMaximalComponentFor(Sequence sequence)
			throws Exception {
		// TODO Auto-generated method stub
		return 0;
	}
	
	public void setTextOutput(boolean out) {
		textOutput=out;
	}


}
