package de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import de.jstacs.data.AlphabetContainer;
import de.jstacs.io.NonParsableException;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.AbstractVariableStructureModel;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures.ParsimoniousElement;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures.ParsimoniousElement.LearningMethod;

/**
 * Abstract class for all trainable statistical models that include parsimonious context trees (PCTs). 
 * @author Ralf Eggeling <eggeling@informatik.uni-halle.de>
 *
 */
public abstract class AbstractParsimoniousModel extends AbstractVariableStructureModel {

	protected ParsimoniousElement[] tree;
	
	protected ParsimoniousElement.ModelType usedModelType = ParsimoniousElement.ModelType.PARSMM;
	
	protected ParsimoniousElement.StructureScore usedStructureScore =  ParsimoniousElement.StructureScore.BD;
	
	protected ParsimoniousElement.ParameterEstimate usedParameterEstimate = ParsimoniousElement.ParameterEstimate.MP;
	
	protected ParsimoniousElement.LearningMethod learningMethod = ParsimoniousElement.LearningMethod.MAXIMIZATION;
		
	protected boolean inSamplingMode = false;
	
	protected boolean trained;
	
	protected String[] parameterSamples;
	
	public AbstractParsimoniousModel(AlphabetContainer alphabets, int length) {
		super(alphabets, length);
	}
	
	public AbstractParsimoniousModel(StringBuffer sb) throws NonParsableException {
		super(sb);
	}
	
	public void setStructureScore(ParsimoniousElement.StructureScore structureScore) {
		usedStructureScore = structureScore;
	}
	
	public void setParameterEstimate( ParsimoniousElement.ParameterEstimate parameterEstimate) {
		usedParameterEstimate = parameterEstimate;
	}
	
	public void setModelType(ParsimoniousElement.ModelType mt) {
		this.usedModelType=mt;
	}
	
	public String getGraphViz() {
		StringBuffer out = new StringBuffer();
		out.append("digraph{\n");
		out.append("edge [arrowsize=0.8 headport=n labelangle=0];\n");
		out.append("node  [label=\"\" width=1.2 height=0.55 color=black,fixedsize=true,fontsize=14];\n");
		for(int i=0;i<tree.length;i++) {
			out.append("/* PCT representation of position "+i+"*/\n");
			out.append(tree[i].getGraphviz());
		}
		out.append("}");
		return out.toString();
	}
	
	public String getTreeStructures() {
		StringBuffer res = new StringBuffer() ;
			for(int i=0;i<tree.length;i++) {
				res.append(tree[i].getTreeStructure());
				res.append('\n');
			}
			
		return res.toString();
	}	
	
	public String getSparseParameterRepresentation() {
		String res = "";
		for(int i=0;i<tree.length;i++) {
			res+=(tree[i].getSparseParameterRepresentation() + "&");
		}
		return res;
	}
	
		
	public void initForSampling(String s) throws IOException {
		sampleFilePath = s;
		writer = new BufferedWriter( new FileWriter( new File(s), true ) );
		inSamplingMode = true;
		this.learningMethod=LearningMethod.SAMPLING;
	}
	
	public void saveCurrentParametersToFile() throws IOException {
		writer.append(this.getSparseParameterRepresentation()+"\n");
		writer.flush();
		numberOfSamples++;
	}
	
	
	public void samplingStopped() throws IOException {
		if( writer != null ) {
			writer.close();
			writer = null;
		}
		inSamplingMode = false;
	}
	
	
	public ParsimoniousElement.LearningMethod getLearningMethod() {
		return learningMethod;
	}
	
	public void parse(String str) {
		String[] s = str.substring(str.indexOf( "\t" )+1).split("&");
		for(int i=0;i<tree.length;i++) {
			tree[i].loadParametersFromSparseRepresentation(s[i]);
		}
		trained=true;
	}
	
	public boolean isInitialized() {
		return trained;
	}
	
}