package de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeSet;
import java.util.regex.Pattern;


import de.jstacs.io.NonParsableException;
import de.jstacs.NotTrainedException;
import de.jstacs.Storable;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongAlphabetException;
import de.jstacs.data.AlphabetContainer;
import de.jstacs.data.WrongLengthException;
import de.jstacs.data.alphabets.DiscreteAlphabet;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.XMLParser;
import de.jstacs.sequenceScores.statisticalModels.trainable.discrete.inhomogeneous.SequenceIterator;
import de.jstacs.utils.Normalisation;
import de.jstacs.utils.functions.Bayesian;
import de.jstacs.utils.functions.NML;
import de.jstacs.utils.random.DirichletMRG;
import de.jstacs.utils.random.DirichletMRGParams;
import de.jtem.numericalMethods.calculus.specialFunctions.Gamma;

public class ParsimoniousElement implements Cloneable, Storable {
	
	public static enum StructureScore {
		BD, FNML, AIC, BIC, DIRONEHALF;
		
		public static StructureScore parseString(String s) {
			if(s.equals("BD")||s.equals("0"))
				return BD;
			if(s.equals("FNML")||s.equals("1"))
				return FNML;
			if(s.equals("AIC")||s.equals("2"))
				return AIC;
			if(s.equals("BIC")||s.equals("3"))
				return BIC;
			if(s.equals("DIRONEHALF")||s.equals("4"))
				return DIRONEHALF;
			return BD;
		}
	}
	
	public static enum ParameterEstimate {
		MP, FSNML, DIRONEHALF, SAMPLING, ML;
		
		public static ParameterEstimate parseString(String s) {
			if(s.equals("MP")||s.equals("0"))
				return MP;
			if(s.equals("FSNML")||s.equals("1"))
				return FSNML;
			if(s.equals("DIRONEHALF")||s.equals("2"))
				return DIRONEHALF;
			if(s.equals("SAMPLING")||s.equals("3"))
				return SAMPLING;
			if(s.equals("ML")||s.equals("4"))
				return ML;
			return MP;
		}
	}
	
	public static enum LearningMethod {
		SAMPLING, MAXIMIZATION;
		
		public static LearningMethod parseString(String s) {
			if(s.equals("SAMPLING")||s.equals("0"))
				return SAMPLING;
			if(s.equals("MAXIMIZATION")||s.equals("1"))
				return MAXIMIZATION;
			return MAXIMIZATION;
		}
	}
	
	public static enum ModelType {
		PARSMM, VOMM, FOMM;
		
		public static ModelType parseString(String s) {
			if(s.equals("PARSMM")||s.equals("0"))
				return PARSMM;
			if(s.equals("VOMM")||s.equals("1"))
				return VOMM;
			if(s.equals("FOMM")||s.equals("2"))
				return FOMM;
			return PARSMM;
		}
	}
	
	private final String XML_TAG = "ParsimoniousElement";
	
	/**
	 * the root of the context tree
	 */
	
	public ContextTreeNode root;
	
	/**
	 * the maximal depth of the context tree
	 */
	private byte maximalDepth;
	
	/**
	 * the equivalent sample size of the model
	 */
	private double ESS;
	
	/**
	 * the constant of the structure prior of the context tree
	 */
	private double treePriorConstant;
	
	/**
	 * the alphabet container of the full model
	 */
	private AlphabetContainer con;
	
	/**
	 * probabilityParameters[a][b] is the probability of observing symbol b given context a. 
	 */
	private double[][] probabilityParameters;
	
	/**
	 * the regular expressions of all leaves of the context tree 
	 */
	private String[] regEx;
	
	/**
	 * partition sets for all layers of the context trees 
	 */
	private PartitionSet[] partset;
	
	/**
	 * contains the absolute frequency of observing a symbol with a given context
	 */
	private double[][] markovCounts;
	
	/**
	 * the sizes of the alphabets of the context positions
	 */
	private int[] contextAlphabetLength;
	
	/**
	 * structure of the current context tree as string
	 */
	private String treeStructure;
	
	/**
	 * structure of the current context tree as graphViz representation
	 */
	private String graphViz;
	
	/**
	 * next leaf to be visited during tree traversal
	 */
	private int nextLeaf;
	
	/**
	 * next node to be visited during tree traversal
	 */
	private int nextNode;
	
	
	/**
	 * true if the cond. prob. parameters are filled
	 */
	private boolean trained;
	
	/**
	 * 
	 */
	private int[] blockSize;
	
	/**
	 * 
	 */
	private int sampleSize;
	
	/**
	 * the number of the parsimoniousElement in a wider context, e.g. as part of an InhomogeneousMarkovModel.
	 * The elementNumber indicates the random variable in the model that this ParsimoniousElement is responsible for.
	 */
	private int elementNumber;
	
	/**
	 * the data set used for training
	 */
	protected DataSet data;
	
	
	public ModelType mt = ModelType.PARSMM;
	
	/**
	 * Main constructor
	 * @param con The alphabet container
	 * @param maximalDepth The maximal depth of the context tree
	 * @param ESS the equivalent sample size
	 * @param treePriorConstant the constant of the structure prior
	 * @throws Exception if something went wrong
	 */
	public ParsimoniousElement(AlphabetContainer con, byte maximalDepth, double ESS, double treePriorConstant) throws Exception {
		if((!con.isSimple())&&con.getNumberOfAlphabets()!=maximalDepth+1)
			throw new WrongAlphabetException("The length of the alphabet container does not correspond to the depth of the context tree.");
		this.con = con;
		this.ESS = ESS;
		this.maximalDepth = maximalDepth;
		this.treePriorConstant = treePriorConstant;
		
		contextAlphabetLength = new int[maximalDepth];
		for(int i=0;i<contextAlphabetLength.length;i++) {
			contextAlphabetLength[i] = (int) con.getAlphabetLengthAt(i);
		}
		partset = new PartitionSet[maximalDepth];
		int totalNumberOfPossibleSequences = 1;
		for(int i=0;i<partset.length;i++) {
			partset[i] = new PartitionSet(SymbolSet.createSymbolSets((DiscreteAlphabet) con.getAlphabetAt(i)));
			totalNumberOfPossibleSequences *= con.getAlphabetAt(i).length() ;
		}
		markovCounts = new double[totalNumberOfPossibleSequences][(int) con.getAlphabetAt(maximalDepth).length()];
		if(maximalDepth>0) {
			blockSize = new int[maximalDepth];
			blockSize[0] = (int) (markovCounts.length/contextAlphabetLength[0]);
			for(int i=1;i<blockSize.length;i++) {
				blockSize[i] = (int) (blockSize[i-1]/contextAlphabetLength[i]);
			}
		}
	}
	
	public double getScore() {
		return root.getLogScore();
	}
	
	/**
	 * Loading a ParsimoniousElement (with or without drawn parameters) from a StringBuffer. 
	 * @param xml the XML-representation
	 * @throws Exception if something went wrong
	 */
	public ParsimoniousElement(StringBuffer xml) throws Exception {
		StringBuffer buff = XMLParser.extractForTag(xml, XML_TAG);
		this.con = XMLParser.extractObjectForTags(buff, "AlphabetContainer",AlphabetContainer.class);
		this.ESS = XMLParser.extractObjectForTags(buff, "ESS",Double.class);
		this.maximalDepth = XMLParser.extractObjectForTags(buff, "maximalDepth",Byte.class);
		this.treePriorConstant = XMLParser.extractObjectForTags(buff, "treePriorConstant",Double.class);
		this.trained = XMLParser.extractObjectForTags(buff, "trained",Boolean.class);
		if(trained) {
			this.treeStructure = XMLParser.extractObjectForTags(buff, "treeStructure", String.class);
			this.graphViz = XMLParser.extractObjectForTags(buff, "graphViz", String.class);
			this.regEx = XMLParser.extractObjectForTags(buff, "regularExpressions", String[].class);
			this.probabilityParameters = XMLParser.extractObjectForTags(buff, "probabilityParameters", double[][].class);
		}
		contextAlphabetLength = new int[maximalDepth];
		for(int i=0;i<contextAlphabetLength.length;i++) {
			contextAlphabetLength[i] = (int) con.getAlphabetLengthAt(i);
		}
		partset = new PartitionSet[maximalDepth];
		int totalNumberOfPossibleSequences = 1;
		for(int i=0;i<partset.length;i++) {
			partset[i] = new PartitionSet(SymbolSet.createSymbolSets((DiscreteAlphabet) con.getAlphabetAt(i)));
			totalNumberOfPossibleSequences *= con.getAlphabetAt(i).length() ;
		}
		markovCounts = new double[totalNumberOfPossibleSequences][(int) con.getAlphabetAt(maximalDepth).length()];
		if(maximalDepth>0) {
			blockSize = new int[maximalDepth];
			blockSize[0] = (int) (markovCounts.length/contextAlphabetLength[0]);
			for(int i=1;i<blockSize.length;i++) {
				blockSize[i] = (int) (blockSize[i-1]/contextAlphabetLength[i]);
			}
		}
	}
	
	public ParsimoniousElement clone() throws CloneNotSupportedException {
		ParsimoniousElement cl = (ParsimoniousElement) super.clone();
		if(probabilityParameters!=null)
		cl.probabilityParameters = probabilityParameters.clone();
		if(regEx!=null)
		cl.regEx =regEx.clone();
		if(root!=null)
		cl.root = root.clone();
		return cl;
	}
	
	public void setElementNumber(int n) {
		this.elementNumber=n;
	}
	
	/**
	 * Samples a new context tree structure and corresponding probability parameters given (weighted) input data
	 * @param data
	 * @param weights
	 * @param contextPositions
	 * @return returns the score of the sampled tree
	 * @throws Exception 
	 */
	public double sampleStructureAndParameters(DataSet data, double[] weights) throws Exception {
		sampleStructure( data, weights);
		regEx = new String[root.getNumberOfLeaves()];
		nextLeaf = 0;
		graphViz = "";
		root.traverseAfterConstruction(ParameterEstimate.SAMPLING);
		trained = true;
		return root.logScore;
	}
	
	private void initializeForNewPruningOperation(DataSet data, double[] weights) throws Exception {
		sampleSize = data.getNumberOfElements();
		this.data=data;
		resetMarkovCounts();
		treeStructure = new String();
		if(data!=null) {
			if(weights!=null) {
	//			computeMarkovCounts(data,weights);
			}
			else {
				double[] localWeights = new double[data.getNumberOfElements()];
				for(int i=0;i<localWeights.length;i++) {
					localWeights[i] = 1;
				}
	//			computeMarkovCounts(data,localWeights);
			}
		}
	}
	
	public double sampleStructure(DataSet data, double[] weights) throws Exception {
		this.initializeForNewPruningOperation(data, weights);
		root = new ContextTreeNode(LearningMethod.SAMPLING, StructureScore.BD);
		probabilityParameters = new double[root.getNumberOfLeaves()][(int) con.getAlphabetLengthAt(maximalDepth)];
		return root.logScore;
	}
	
	public void maximizeStructureAndParameters(DataSet data, double[] weights, StructureScore structureScore, ParameterEstimate parameterEstimate) throws Exception {
		this.initializeForNewPruningOperation(data, weights);
		this.data=data;
		root = new ContextTreeNode(LearningMethod.MAXIMIZATION, structureScore);
		probabilityParameters = new double[root.getNumberOfLeaves()][(int) con.getAlphabetLengthAt(maximalDepth)];
		regEx = new String[root.getNumberOfLeaves()];
		nextLeaf = 0;
		graphViz = "";
		nextNode = 0;
		root.traverseAfterConstruction(parameterEstimate);
		trained = true;
	}

	/**
	 * Computes Markov counts (e.g. conditional nucleotide frequencies) from data
	 * @param data
	 * @param weights
	 * @throws WrongAlphabetException
	 * @throws WrongLengthException
	 */
/*	public void computeMarkovCounts(DataSet data, double[] weights) throws WrongAlphabetException, WrongLengthException {
		resetMarkovCounts();
		if(!data.getAlphabetContainer().checkConsistency(con))
			throw new WrongAlphabetException();
		if(data.getElementLength()!=(maximalDepth+1))
			throw new WrongLengthException("Sample size does not equal maximal tree size + 1 ("+data.getElementLength()+"!="+(maximalDepth+1)+")");
		if(maximalDepth>0) {
			for(int i=0;i<data.getNumberOfElements();i++) {
	/*			int seqId = 0;
				Sequence seq = data.getElementAt(i);
				for(int j=0;j<seq.getLength()-1;j++) {
					seqId += seq.discreteVal(j)*blockSize[j];
				}*/
	/*			markovCounts[getContextIDForSequence(data.getElementAt(i))][data.getElementAt(i).discreteVal(data.getElementAt(i).getLength()-1)]+=weights[i];						
			}
		} else {
			for(int i=0;i<data.getNumberOfElements();i++) {
				markovCounts[0][data.getElementAt(i).discreteVal(0)]+=weights[i];						
			}
		}
	} */
	
	public void addSequenceToMarkovCounts(Sequence seq) {
		if(maximalDepth>0) {
			markovCounts[getContextIDForSequence(seq)][seq.discreteVal(seq.getLength()-1)]++;						
		} else {
			markovCounts[0][seq.discreteVal(0)]++;						
		}
	}
	
	public void removeSequenceFromMarkovCounts(Sequence seq) {
		if(maximalDepth>0) {
			markovCounts[getContextIDForSequence(seq)][seq.discreteVal(seq.getLength()-1)]--;						
		} else {
			markovCounts[0][seq.discreteVal(0)]--;						
		}
	}
	
	private int getContextIDForSequence(Sequence seq) {
		int seqId = 0;
		for(int j=0;j<seq.getLength()-1;j++) {
			seqId += seq.discreteVal(j)*blockSize[j];
		}
		return seqId;
	}
	
	/**
	 * Computes the probability of data (represented by Markov counts) given the context tree
	 * @param counts
	 */
	public double getLogProbForCurrentMarkovCounts() {
		return root.getLogProbForCurrentMarkovCounts();
	}
	
	/**
	 * 
	 * @return the number of leaves / the number of conditional probability distributions of this ParsimoniousElement as as a measure for statistical complexity
	 */
	public int getNumberOfLeaves() {
		return probabilityParameters.length;
	}
	
	private void resetMarkovCounts() {
		for(int i=0;i<markovCounts.length;i++) {
			for(int j=0;j<markovCounts[i].length;j++)
				markovCounts[i][j] = 0;
		}
	}
	
	
	/**
	 * 
	 * @param seq An arbitrary discrete sequence
	 * @param position the position of the symbol whose probability is to be computed
	 * @return returns the probability of the symbol at the given positions given the current context tree, given the precedent observations, given the probability parameters 
	 * @throws NotTrainedException
	 */
	public double getLogProbFor(Sequence seq, int position) throws NotTrainedException {
		if(!trained)
			throw new NotTrainedException();
		if(root!=null)
			return Math.log(root.getProbFor(seq, position));
		else {
			double res = 0;
			String str = seq.getSubSequence(position-maximalDepth,maximalDepth).toString();
			for(int i=0;i<regEx.length;i++) {
				if(Pattern.matches(regEx[i],str)) {
					res = Math.log(this.probabilityParameters[i][seq.discreteVal(position)]);
					break;
				}
			}
			return res;
		}
	}
	
/*	public double getLogBDScoreForGivenMarkovCounts(double[][] markovCounts) throws NotTrainedException, WrongLengthException, WrongAlphabetException {
		if(!trained)
			throw new NotTrainedException();
		resetMarkovCounts();
		this.markovCounts = markovCounts.clone();
		return root.getLogBDScore();
	}
	*/
	
	
	/**
	 * A node is represented by its label, surrounded by [ ] brackets, followed by the representation of all children in ( ) brackets.
	 * The root does not have a label, hence each context tree is starts with ( and ends with ).  
	 * @return the structure of the context tree as bracket representation. 
	 * 
	 * 
	 *  */
	public String getTreeStructure() {
		if(maximalDepth==0)
			return "()";
		else
			return "("+treeStructure.toString()+")";
	}
	
	/**
	 * 
	 * @return A sparse representation of the parameters to be stored in a file during sampling.
	 */
	public String getSparseParameterRepresentation() {
		StringBuffer res = new StringBuffer();
		for(int i=0;i<probabilityParameters.length;i++) {
			res.append(regEx[i]);
			res.append("\t");
			for(int j=0;j<probabilityParameters[i].length;j++) {
				res.append(probabilityParameters[i][j]);
				res.append("\t");
			}
			res.append("%");
		}
		return res.toString();
	}
	
	/** 
	 * returns the graphViz representation of this context tree
	 */
	public String getGraphviz(){
		return graphViz;
	}
	
	/**
	 * 
	 * @param str sparse parameter representation
	 */
	public void loadParametersFromSparseRepresentation(String str) {
		String[] s = str.split("%");
		probabilityParameters = new double[s.length][(int) con.getAlphabetLengthAt(maximalDepth)];
		regEx = new String[s.length];
		for(int i=0;i<s.length;i++) {
			String[] d = s[i].split("\t");
			if(d.length== con.getAlphabetLengthAt(maximalDepth)) {
				regEx[i]="";
				for(int j=0;j<probabilityParameters[i].length;j++) {
					probabilityParameters[i][j] = Double.parseDouble(d[j]);
				}
			} else {
				regEx[i]=d[0];
				for(int j=0;j<probabilityParameters[i].length;j++) {
					probabilityParameters[i][j] = Double.parseDouble(d[j+1]);
				}
			}
			
		}
		trained=true;
		root = null;
	}
	
	/**
	 * 
	 * @author eggeling
	 * Represents a node of a parsimonious context tree.
	 */
	
	private class ContextTreeNode implements Comparable<ContextTreeNode>, Cloneable {
		
		/**
		 * the distance of this node to the root
		 */
		private int distanceToRoot;
		
		/**
		 * the label of the node, which is a subset of all possible symbols at this layer of the context tree
		 */
		private final SymbolSet label;
		
		/**
		 * the regular expression encoded by this node
		 */
		private String regExStr;
		
		/**
		 * the ESS of the subtree rooted under this node
		 */
		private double subtreeESS;
		
		/**
		 * the number of leaves of the subtree under this node
		 */
		private int numberOfLeaves = 0;
		
		/**
		 * the counts in this node. only filled, if node is leaf.
		 */
		private double[] counts;
		
		/**
		 * the hyperparameters in this node. only filled, if node is leaf.
		 */
		private double[] hyperparameters;
		
		/**
		 * the log score of the context tree under this leaf
		 */
		private double logScore;
		
		/**
		 * the log of the sum of all possible context trees under this leaf
		 */
//		private double logSumScore;
		
		/**
		 * the set of children nodes, null if node is leaf
		 */
		private TreeSet<ContextTreeNode> children;
		
		/**
		 * the index of the leaf corresponding to the order of being visited during tree traversal
		 */
		private int leafIndex;
		
		/**
		 * The number of the node in the final context tree
		 */
		protected int nodeNumber = 0;
		
		/**
		 *
		 */
		protected TreeSet<Integer> representedSequenceIDs;

		/**
		 * Constructor for the root of a context tree
		 */
		protected ContextTreeNode(LearningMethod method, StructureScore typeOfScore) {
			distanceToRoot = 0;
			label = null;
			regExStr = "";
			subtreeESS = ESS;
			representedSequenceIDs = new TreeSet<Integer>();
			for(int i=0;i<sampleSize;i++) {
				representedSequenceIDs.add(i);
			} 
			if(maximalDepth>0) {
				createChildren(method,typeOfScore);
				if(method==LearningMethod.SAMPLING) {
					pruneBySampling();
					} else if(method==LearningMethod.MAXIMIZATION) {
						pruneByMaximization();
						} 
				} else {
					numberOfLeaves = 1;
					if(typeOfScore==StructureScore.BD){
						logScore = computeLogBDScoreOfLeaf();
						} else if(typeOfScore==StructureScore.FNML) {
							logScore = computeLogfNMLScoreOfLeaf();
							} else if(typeOfScore==StructureScore.DIRONEHALF) {
								logScore = computeLogDIRONEHALFScoreOfLeaf();
								} else if(typeOfScore==StructureScore.BIC) {
									logScore = this.computeLogBICScoreOfLeaf();
									} else if(typeOfScore==StructureScore.AIC) {
										logScore = this.computeLogAICScoreOfLeaf();
									}
					}
			}
		
		/**
		 * Creates an inner node or a leaf
		 * @param parent
		 * @param label
		 */
		protected ContextTreeNode(ContextTreeNode parent, SymbolSet label, LearningMethod method, StructureScore typeOfScore) {
			this.label = label;
			subtreeESS = parent.getSubtreeESS()/label.getAlphabet().length()*label.getNumberOfElements();
			distanceToRoot = parent.getDistanceToRoot() + 1;
			String str = "[";
			for(int i=0;i<label.getNumberOfElements();i++) {
				str+=label.getElements()[i];
			}
			str+="]";
			regExStr = str+parent.getRegExStr();
			representedSequenceIDs = new TreeSet<Integer>();
			Iterator<Integer> it = parent.representedSequenceIDs.iterator();
			while(it.hasNext()) {
				int id = it.next();
				if(label.contains(data.getAlphabetContainer().getSymbol(0, data.getElementAt(id).discreteVal(data.getElementAt(id).getLength()-1-this.distanceToRoot))))
					representedSequenceIDs.add(id);
			} 
	//		System.out.println(regExStr+"\t"+representedSequenceIDs.size());
	//		if(representedSequenceIDs.size()==0) {
	//			numberOfLeaves = 1;
	//			logScore = Double.NEGATIVE_INFINITY;
	//		} else
			{
			if(distanceToRoot<maximalDepth) {
				createChildren(method,typeOfScore);
			} else {
				numberOfLeaves = 1;
		//		logScore = computeLogBDScoreOfLeaf(true);
				if(typeOfScore==StructureScore.BD){
					logScore = computeLogBDScoreOfLeaf();
					} else if(typeOfScore==StructureScore.FNML) {
						logScore = computeLogfNMLScoreOfLeaf();
					} else if(typeOfScore==StructureScore.DIRONEHALF) {
						logScore = computeLogDIRONEHALFScoreOfLeaf();
					} else if(typeOfScore==StructureScore.BIC) {
						logScore = this.computeLogBICScoreOfLeaf();
					} else if(typeOfScore==StructureScore.AIC) {
						logScore = this.computeLogAICScoreOfLeaf();
					}
		//		System.out.println(regExStr + "\t" + logScore + "\t" + Math.exp(logScore));
			}
			}
		}
		
		public ContextTreeNode clone() throws CloneNotSupportedException {
			ContextTreeNode cl = (ContextTreeNode)super.clone();
	//		cl.children = (TreeSet<ContextTreeNode>) children.clone();
	//		cl.counts = counts.clone();
	//		cl.hyperparameters = hyperparameters.clone();
	//		cl.label = label.clone();
			return cl;
		}
		
		/**
		 * creates all possible children of a node, i.e. one child for each possible SymbolSet
		 */
		private void createChildren(LearningMethod method, StructureScore typeOfScore) {
			children = new TreeSet<ContextTreeNode>();
			for(SymbolSet set: partset[partset.length-1-distanceToRoot].getSymbolSet()) {
				ContextTreeNode child = new ContextTreeNode(this,set,method,typeOfScore);
				if(child.getDistanceToRoot()<maximalDepth) {
					if(method==LearningMethod.SAMPLING) {
						child.pruneBySampling();
					} else if(method==LearningMethod.MAXIMIZATION) {
						child.pruneByMaximization(); 
						}
				}
				children.add(child);
			}
		}
		
		/**
		 * Reduces the children to a valid partition by sampling a valid partition from the space of all valid partition according to the score.
		 */
		protected void pruneBySampling() {
			// compute the score for each possible partition
			Partition[] partarr = partset[partset.length-1-distanceToRoot].getPartition().toArray(new Partition[partset[partset.length-1-distanceToRoot].getPartition().size()]);
			for(int i=0;i<partarr.length;i++) {
				double localscore = 0;
				for(ContextTreeNode ctn: this.children) {
					if(partarr[i].contains(ctn.getLabel())) {
						localscore+= ctn.getLogScore();
					}
				}
				partarr[i].setScore(localscore);
			}
			
			double[] logScoreArr = new double[partarr.length];
			for(int i=0;i<logScoreArr.length;i++) {
				logScoreArr[i] = partarr[i].getScore();
			}
			Normalisation.logSumNormalisation(logScoreArr);
			for(int i=0;i<partarr.length;i++) {
				partarr[i].setScore(logScoreArr[i]);
			}
			
			// draw a partition
			double r = new Random().nextDouble();
			boolean finished = false;
			double probSum = 0;
			int drawnPartIndex = 0;
			for(int i=0;i<partarr.length&&(!finished);i++) {
				probSum+=partarr[i].getScore();
				if(r<probSum) {
					drawnPartIndex = i;
					finished = true;
				}
			}
			Partition drawnPartition = partarr[drawnPartIndex];
			
			// remove all children not belonging to the drawn partition
			TreeSet<ContextTreeNode> toBeRemoved = new TreeSet<ContextTreeNode>();
			for(ContextTreeNode tau : this.children) {
				if(!drawnPartition.contains(tau.getLabel())) {
					toBeRemoved.add(tau);
				}
			}
			for(ContextTreeNode tau : toBeRemoved) {
				children.remove(tau);
			}
			
			// update score and leaf number of this node 
			for(ContextTreeNode tau: this.children) {
				this.logScore += tau.getLogScore();
				this.numberOfLeaves += tau.getNumberOfLeaves();
			}
		}
		
		/**
		 * Prunes by maximization of local BD scores. For comparison with sampling.
		 */
		protected void pruneByMaximization() {
			try{
			
			Partition[] partarr = partset[partset.length-1-distanceToRoot].getPartition().toArray(new Partition[partset[partset.length-1-distanceToRoot].getPartition().size()]);
			for(int i=0;i<partarr.length;i++) {
				double localscore = 0;
				for(ContextTreeNode ctn: this.children) {
					if(partarr[i].contains(ctn.getLabel())) {
						localscore+= ctn.getLogScore();
					}
				}
				partarr[i].setScore(localscore);
			}
			
			double[] logScoreArr = new double[partarr.length];
			for(int i=0;i<logScoreArr.length;i++) {
				logScoreArr[i] = partarr[i].getScore();
			}

			double maximalScore = Double.NEGATIVE_INFINITY;
			int maxPartIndex = 0;
			
			if(mt==ModelType.VOMM) {
				if(this.label==null||this.label.getNumberOfElements()==1) {
					for(int i=0;i<partarr.length;i++) {
						if(i<11||i>13)
								if(partarr[i].getScore()>maximalScore) {
									maximalScore = partarr[i].getScore();
									maxPartIndex = i;
								} 
							}
				} else {
					maxPartIndex = partarr.length-1;
				}
			}
			if(mt==ModelType.PARSMM) {
				for(int i=0;i<partarr.length;i++) {
			//		if(elementNumber==4)
			//		System.out.print(partarr[i].toString()+"\t"+partarr[i].getScore());
					if(partarr[i].getScore()>maximalScore) {
			//			if(elementNumber==4)
			//			System.out.println("is maximal");
						maximalScore = partarr[i].getScore();
						maxPartIndex = i;
					} else {
			//			if(elementNumber==4)
			//			System.out.println();
					}
				} 
			}
			if(mt==ModelType.FOMM) {
				for(int i=0;i<partarr.length;i++) {
					if(i==0||i==partarr.length-1)
						if(partarr[i].getScore()>maximalScore) {
							maximalScore = partarr[i].getScore();
							maxPartIndex = i;
						} 
				} 
			}
			
			Partition drawnPartition = partarr[maxPartIndex];
	//		System.out.println(this.label+"\t"+maximalScore+"\t"+drawnPartition.toString());
	//		System.out.println("selected="+maxPartIndex+"\t"+drawnPartition.toString()+"\t"+drawnPartition.getScore());
			
	  			TreeSet<ContextTreeNode> toBeRemoved = new TreeSet<ContextTreeNode>();
			for(ContextTreeNode tau : this.children) {
				if(!drawnPartition.contains(tau.getLabel())) {
					toBeRemoved.add(tau);
				}
			}
			for(ContextTreeNode tau : toBeRemoved) {
				children.remove(tau);
			}
			
			// update score and leaf number of this node 
			for(ContextTreeNode tau: this.children) {
				this.logScore += tau.getLogScore();
				this.numberOfLeaves += tau.getNumberOfLeaves();
			}
			
		//	System.out.println("atError"+this.nodeNumber);
			/*
			TreeSet<SymbolSet> set = new TreeSet<SymbolSet>();
			for(ContextTreeNode ctn: this.children) {
				SymbolSet s = ctn.getLabel();
				s.setScore(ctn.getLogScore());
				set.add(s);
			}		
			double[] gScore = new double[set.size()];
			Arrays.fill(gScore, 0);
			SymbolSet[] ssarr = new SymbolSet[set.size()];
			HashMap<String,SymbolSet> hm = new HashMap<String,SymbolSet>();
			int j=0;
			Iterator<SymbolSet> it = set.iterator();
			while(it.hasNext()) {
				SymbolSet s = it.next();
				hm.put(s.toSparseString(), s);
				ssarr[j] = s;
				s.ID=j;
				j++;
			}
			for(int i=0;i<ssarr.length;i++) {
				gScore[i] = getGScore(ssarr[i],set,hm,gScore);
			}
			TreeSet<Integer> tsi = ssarr[ssarr.length-1].optimalSubSetIDs;
			TreeSet<SymbolSet> tobeRetained = new TreeSet<SymbolSet>();
			while(!tsi.isEmpty()) {
				SymbolSet s = ssarr[tsi.pollFirst()];
				tobeRetained.add(s);
			}
			
			// remove all children not belonging to the maximal
			TreeSet<ContextTreeNode> toBeRemoved = new TreeSet<ContextTreeNode>();
			for(ContextTreeNode tau : this.children) {
				if(!tobeRetained.contains(tau.getLabel())) {
					toBeRemoved.add(tau);
				}
			}
			for(ContextTreeNode tau : toBeRemoved) {
				children.remove(tau);
			}
			
			// update score and leaf number of this node 
			for(ContextTreeNode tau: this.children) {
				this.logScore += tau.getLogScore();
				this.numberOfLeaves += tau.getNumberOfLeaves();
			} */
			}catch(Exception e) { // System.out.println(this.distanceToRoot+"\t"+this.label); 		
				e.printStackTrace();
			}
		}
		
		protected double[] computeCounts() {
			double[] lcounts = new double[(int)con.getAlphabetAt(con.getNumberOfAlphabets()-1).length()];
			Iterator<Integer> it = representedSequenceIDs.iterator();
			while(it.hasNext()) {
				int next = it.next();
				lcounts[data.getElementAt(next).discreteVal(data.getElementAt(next).getLength()-1)]++;
			}
			return lcounts;
		}
		
		/**
		 * 
		 * @return The BD-score of a leaf, that is B(\vec{N}+\vec{\alpha})/B(\vec{alpha}), where B denotes the multinomial Beta function.
		 */
		
		protected double computeLogBDScoreOfLeaf() {
			counts = computeCounts();
			fillHyperparametersVector();
			return Bayesian.logOfBayesFactor(counts,hyperparameters)+treePriorConstant;
		}
		
		
		/**
		 * 
		 * @return The Bayes score of a leaf with hyperparameters set to 1/2 (Jeffrey's prior)
		 */
		protected double computeLogDIRONEHALFScoreOfLeaf() {
			counts = computeCounts();
			fillHyperparametersVectorWithOneHalf();
			return Bayesian.logOfBayesFactor(counts,hyperparameters)+treePriorConstant;
		}
		
		/**
		 * 
		 * @return the fNML score of a particular leaf
		 */
		protected double computeLogfNMLScoreOfLeaf() {
			counts = computeCounts();
			double localN = 0;
			for(int i=0;i<counts.length;i++) {
				localN += counts[i];
			}
			double logLikelihoodTerm = 0;
			for(int i=0;i<counts.length;i++) {
				if(counts[i]>0)
					logLikelihoodTerm += counts[i]*(Math.log(counts[i])-Math.log(localN));
			}
			return logLikelihoodTerm - NML.SzpankowskiApproximation(counts.length, (int)localN) + treePriorConstant;
		}
		
		/**
		 * 
		 * @return the BIC score of the leaf
		 */
		protected double computeLogBICScoreOfLeaf() {
			counts = computeCounts();	
			double localN = 0;
			for(int i=0;i<counts.length;i++) {
				localN += counts[i];
			}
			double logLikelihoodTerm = 0;
			for(int i=0;i<counts.length;i++) {
				if(counts[i]>0)
					logLikelihoodTerm += counts[i]*(Math.log(counts[i])-Math.log(localN));
			}
			return 2*logLikelihoodTerm - (counts.length-1)*Math.log(sampleSize);
		}
		
		/**
		 * 
		 * @return the AIC score of the leaf
		 */
		protected double computeLogAICScoreOfLeaf() {
			counts = computeCounts();
			double localN = 0;
			for(int i=0;i<counts.length;i++) {
				localN += counts[i];
			}
			double logLikelihoodTerm = 0;
			for(int i=0;i<counts.length;i++) {
				if(counts[i]>0)
					logLikelihoodTerm += counts[i]*(Math.log(counts[i])-Math.log(localN));
			}
			return 2*logLikelihoodTerm - 2*(counts.length-1);
		} 
		
		
		
		
		protected double getLogProbForCurrentMarkovCounts() {
			double res = 0;
			if(this.children!=null) {
				for(ContextTreeNode tau: this.children) {
					res+=tau.getLogProbForCurrentMarkovCounts();
				}
			} else {
				res = computeLogBDScoreOfLeaf();
			}
			return res;
		}
	
		/**
		 * A node A is smaller than a node B if the label/SymbolSet of A is smaller than the label/SymbolSet of B 
		 */
		public int compareTo(ContextTreeNode o) {
			return this.label.compareTo(o.label);
		}
	
		/**
		 * if return value 0 -> node is root
		 * if return value maximalDepth -> node is leaf
		 * @return distance to root
		 */
		protected int getDistanceToRoot() {
			return distanceToRoot;
		}
		
		/**
		 * 
		 * @return the label/SymbolSet of a node
		 */
		protected SymbolSet getLabel() {
			return label;
		}
		
		/**
		 * 
		 * @return the number of leaves of the subtree under this node
		 */
		protected int getNumberOfLeaves() {
			return numberOfLeaves;
		}
		
		private void fillHyperparametersVector() {
			hyperparameters = new double[(int)con.getAlphabetLengthAt(maximalDepth)];
			for(int i=0;i<hyperparameters.length;i++) {
				hyperparameters[i]=subtreeESS/hyperparameters.length;
			}
		}
		
		private void fillHyperparametersVectorWithOneHalf() {
			hyperparameters = new double[(int)con.getAlphabetLengthAt(maximalDepth)];
			for(int i=0;i<hyperparameters.length;i++) {
				hyperparameters[i]=0.5;
			}
		}
		
		private double[] getSumOfCountsAndHyperparameters() {
			double[] res = new double[counts.length];
			for(int i=0;i<res.length;i++) {
				res[i] = counts[i]+hyperparameters[i];
			}
			return res;
		}
		
	/*	protected double getLogBDScore() {
			double res = 0;
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> it = this.children.iterator();
				while(it.hasNext())
					res += it.next().getLogBDScore();
			} else {
				logScore = computeBDScoreOfLeaf();
			}
			return res;
		} */ 
		
		/**
		 * 
		 * @return the logarithmic score (without prior term) of the tree under this node
		 */
		public double getLogScore() {
			return logScore;
		}
		
		/**
		 * 
		 * @return the regular expression encoded by this node
		 */
		protected String getRegExStr() {
			return regExStr;
		}
		
		/**
		 * returns the number of pseudo-counts that are assigned to this subtree		
		 * @return the subtree ess
		 */
		protected double getSubtreeESS() {
			return subtreeESS;
		}
		
		/**
		 * traverses the tree after construction. computes compact tree representation and samples conditional probability parameters. 
		 */
		protected void traverseAfterConstruction(ParameterEstimate parameters) {
			String l = "";
			if(label!=null)
				l=this.label.toString().substring(1,this.label.toString().length()-1);
			graphViz+=("t"+elementNumber+"n"+this.nodeNumber+"[label=\""+l+"\"];\n");
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> iter = children.iterator();
				while(iter.hasNext()) {
					nextNode++;
					ContextTreeNode n = iter.next();
					n.nodeNumber=nextNode;
					graphViz+=("t"+elementNumber+"n"+this.nodeNumber+"-> t"+elementNumber+"n"+n.nodeNumber+";\n");
				}
			}else{
		//		graphViz+=("t"+elementNumber+"n"+this.nodeNumber+"-> t"+elementNumber+"n"+this.nodeNumber+"p;\n");
		//		graphViz+=("t"+elementNumber+"n"+this.nodeNumber+"p[label=\""+(int)this.counts[0]+"\\n"+(int)this.counts[1]+"\\n"+(int)this.counts[2]+"\\n"+(int)this.counts[3]+"\" shape=rectangular height=1.1];\n");
			}
			if(distanceToRoot<maximalDepth) {
				if(label!=null) {
					treeStructure+=(label+"(");
				}
				Iterator<ContextTreeNode> it = children.iterator();
				while(it.hasNext()) {
					it.next().traverseAfterConstruction(parameters);
				}
				if(label!=null)
				treeStructure+=")";
			} else {
				treeStructure+=(label);
				{
					leafIndex = nextLeaf;
					if(parameters==ParameterEstimate.SAMPLING) {
						DirichletMRGParams param = new DirichletMRGParams(this.getSumOfCountsAndHyperparameters());
						probabilityParameters[leafIndex] = DirichletMRG.DEFAULT_INSTANCE.generate((int) con.getAlphabetLengthAt(maximalDepth), param);
					} else if(parameters==ParameterEstimate.FSNML) {
						double sum = 0;
						for(int i=0;i<counts.length;i++) {
							sum += NML.eTerm((int)counts[i])*(counts[i]+1);
						}
						for(int i=0;i<counts.length;i++) {
							probabilityParameters[leafIndex][i] = NML.eTerm((int)counts[i])*(counts[i]+1)/sum;
						}
					} else if(parameters == ParameterEstimate.MP) {
						fillHyperparametersVector();
						double[] arr = this.getSumOfCountsAndHyperparameters();
						double localN = 0;
						for(int i=0;i<arr.length;i++) {
							localN += arr[i];
						}
						for(int i=0;i<arr.length;i++) {
							probabilityParameters[leafIndex][i] = arr[i]/localN;
						}						
					}else if(parameters == ParameterEstimate.ML) {
						hyperparameters = new double[(int)con.getAlphabetLengthAt(maximalDepth)];
						Arrays.fill(hyperparameters, 0);
						double[] arr = this.getSumOfCountsAndHyperparameters();
						double localN = 0;
						for(int i=0;i<arr.length;i++) {
							localN += arr[i];
						}
						for(int i=0;i<arr.length;i++) {
							probabilityParameters[leafIndex][i] = arr[i]/localN;
							if(probabilityParameters[leafIndex][i]==0)
								probabilityParameters[leafIndex][i]=1E-10;
						}						
					} else if(parameters == ParameterEstimate.DIRONEHALF) {
						fillHyperparametersVectorWithOneHalf();
						double[] arr = this.getSumOfCountsAndHyperparameters();
						double localN = 0;
						for(int i=0;i<arr.length;i++) {
							localN += arr[i];
						}
						for(int i=0;i<arr.length;i++) {
							probabilityParameters[leafIndex][i] = arr[i]/localN;
						}
					} 
					regEx[leafIndex] = regExStr;
					nextLeaf++;
				}
				
			}
		}
	
		/*
		 * just a method for diagnostics
		protected void traverse() {
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> it = children.iterator();
				while(it.hasNext()) {
					it.next().traverse();
				}
			} else {
				System.out.print(regExStr+"\t");
				for(int i=0;i<probabilityParameters[leafIndex].length;i++) {
					System.out.print(probabilityParameters[leafIndex][i]+"\t");
				}
				System.out.println();
			}
		}
		*/
		protected double getProbFor(Sequence seq, int position) {
			double res = 0;
			if(distanceToRoot<maximalDepth) {
				Iterator<ContextTreeNode> it = children.iterator();
				while(it.hasNext()) {
					ContextTreeNode child = it.next();
					if(child.label.contains(
					((DiscreteAlphabet)con.getAlphabetAt(maximalDepth-1-distanceToRoot)).getSymbolAt( seq.discreteVal(position-1-distanceToRoot )))) {
						res =  child.getProbFor(seq, position);
						break;
					}
				}
			} else {
				res = probabilityParameters[leafIndex][seq.discreteVal(position)];
			}
			return res;
		}
	}

	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, con, "AlphabetContainer");
		XMLParser.appendObjectWithTags(xml, maximalDepth, "maximalDepth");
		XMLParser.appendObjectWithTags(xml, ESS, "ESS");
		XMLParser.appendObjectWithTags(xml, treePriorConstant, "treePriorConstant");
		XMLParser.appendObjectWithTags(xml, trained, "trained");
		if(trained) {
			XMLParser.appendObjectWithTags(xml, treeStructure, "treeStructure");
			XMLParser.appendObjectWithTags(xml, graphViz, "graphViz");
			XMLParser.appendObjectWithTags(xml, regEx, "regularExpressions");
			XMLParser.appendObjectWithTags(xml, probabilityParameters, "probabilityParameters");
		}
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}
/*	
	public void computeParameters(byte value) {
		computeParameters = value;
	}
*/	
	public byte getMaximalDepth() {
		return maximalDepth;
	}
	
	public void printMarkovCounts() {
		for(int i=0;i<markovCounts.length;i++) {
			for(int j=0;j<markovCounts[i].length;j++) {
				System.out.print(markovCounts[i][j]+"\t");
			}
			System.out.println();
		}
	}
	
/*	public static double getGScore(SymbolSet s, TreeSet<SymbolSet> ts, HashMap<String,SymbolSet> hm, double[] gScore) {
		double res = Double.NEGATIVE_INFINITY;
		Iterator<SymbolSet> it = ts.iterator();
		SymbolSet bestT = ts.first();
		while(it.hasNext()) {
			SymbolSet t = it.next();
			if(t.isSubsetOf(s)) {
		//	if(t.isSubsetOf(s)&&(t.getElements()[0].equals(s.getElements()[0]))) {
				String str = s.getSetDifferenceAsString(t);
				if (str.equals("") ) {
						if(s.getScore()>res) {
							res = t.getScore();
							bestT=t;
						}
					} else {
						if(gScore[hm.get(str).ID]+t.getScore()>res) {
							res = gScore[hm.get(str).ID]+t.getScore();
							bestT=t;
						}
					
				}
			}
			if(t.equals(s))
				break;
		}
		s.optimalSubSetIDs.add(bestT.ID);
		String bestStr = s.getSetDifferenceAsString(bestT);
		if(!bestStr.equals("")) {
			s.optimalSubSetIDs.addAll(hm.get(bestStr).optimalSubSetIDs);
		}		
		return res;	
	} */
	
}