package de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures;
import java.util.Iterator;
import java.util.TreeSet;


import de.jstacs.data.alphabets.DNAAlphabet;
import de.jstacs.data.alphabets.DiscreteAlphabet;

public class SymbolSet implements Comparable<SymbolSet> {
	
	private DiscreteAlphabet alph;
	
	private TreeSet<String> ts;
	
	private double score;
	
	public int ID = 0;
	
	public TreeSet<Integer> optimalSubSetIDs;
	
	/**
	 * Creates arbitrary sets of symbols
	 * @param alph the alphabet
	 * @param strings the symbols that are contained in the set
	 */
	public SymbolSet(DiscreteAlphabet alph, String... strings) {
		if(alph.getClass() == DNAAlphabet.class) {
			try {
				this.alph = new DiscreteAlphabet(true,new String[]{"A","C","G","T"});
			} 	catch(Exception e) {
				e.printStackTrace();
			}
		} else {
			this.alph = alph;
		}
		ts = new TreeSet<String>();
		for(int i=0;i<strings.length;i++) {
			ts.add(strings[i]);
		}
		optimalSubSetIDs = new TreeSet<Integer>();
	}
	
	
	/**
	 * Creates an empty symbol set or one that contains all elements of the alphabet 
	 * @param alph the alphabet
	 * @param filled if true, the set will contain all symbol of the alphabet. 
	 * if false, the set will be empty
	 *
	 */
	public SymbolSet(DiscreteAlphabet alph, boolean filled) {
		if(alph.getClass() == DNAAlphabet.class) {
			try {
				this.alph = new DiscreteAlphabet(true,new String[]{"A","C","G","T"});
			} 	catch(Exception e) {
				e.printStackTrace();
			}
		} else {
			this.alph = alph;
		}
		ts = new TreeSet<String>();
		if(filled) {
			for(int i=0;i<alph.length();i++) {
				ts.add(alph.getSymbolAt(i));
			}
		}
	}
	
	protected boolean isSubsetOf(SymbolSet t) {
		if(this.getNumberOfElements()>t.getNumberOfElements()) {
			return false;
		}
		return t.ts.containsAll(this.ts);
	}
	
	protected String getSetDifferenceAsString(SymbolSet t) {
		if(!t.isSubsetOf(this)) {
			return "";
		}
		else {
			String res = "";
			Iterator<String> it = ts.iterator();
			while(it.hasNext()) {
				String s = it.next();
				if(!t.contains(s)) {
					res+=s;
				}
			}
			return res;
		}
	}
	
	protected boolean equals(SymbolSet t) {
		if(alph.equals(t.getAlphabet())) {
			return ts.equals(t.ts);
		} else {
			return false;
		}
	}
	
	protected String[] getElements() {
		Object[] oarr =  ts.toArray();
		String[] strarr = new String[oarr.length];
		for(int i=0;i<strarr.length;i++) {
			strarr[i] = (String)oarr[i];
		}
		return strarr;
	}
	
	protected void add(String str) {
		ts.add(str);
	}
	
	protected void add(String[] strarr) {
		for(int i=0;i<strarr.length;i++) {
			ts.add(strarr[i]);
		}
	}
	
	
	protected void remove(String str) {
		ts.remove(str);
	}
	
	/**
	 * 
	 * @param t an arbitrary number of SymbolSet Objects
	 * @return a union of all given SymbolSets
	 * @throws Exception if the alphabets do not match
	 */
	protected SymbolSet union(SymbolSet... t) throws Exception {
		SymbolSet res = new SymbolSet(alph, this.getElements());
		for(int i=0;i<t.length;i++) {
			if(!(alph.checkConsistency(t[i].getAlphabet()))) {
				throw new Exception("Union impossible as alphabets do not match");
			}
			res.add(t[i].getElements());
		}
		return res;
	}
	
	protected DiscreteAlphabet getAlphabet() {
		return alph;
	}
	
	public String toString() {
		return ts.toString();
	}

	protected int getNumberOfElements() {
		return ts.size();
	}
	
	/**
	 * An order among SymbolSets is defined according to following rules:
	 * 1. number of elements (a SymbolSet s is smaller than a SymbolSet s' if s contains less elements than s')
	 * 2. comparing the elements is ascending order
	 * Two SymbolSets are equal if and only if they contain the same symbols (not necessarily the same objects)
	 */
	public int compareTo(SymbolSet s)  {
		
		if((int)Math.signum(this.getNumberOfElements()-s.getNumberOfElements())!=0)
			return (int)Math.signum(this.getNumberOfElements()-s.getNumberOfElements());
				
		for(int i=0;i<Math.min(s.getNumberOfElements(),this.getNumberOfElements());i++) {
			if(s.getElements()[i].compareTo(this.getElements()[i])!=0) {
				return (this.getElements()[i].compareTo(s.getElements()[i]));
			}
		}
		return 0;
	}
	

	/**
	 * Checks whether at least one Symbol of t is already contained in the Partition
	 * @param t an arbitrary SymbolSet
	 * @return true if one symbol of t is contained in one SymbolSet of this Partition
	 */
	protected boolean containsAtLeastOneElementOf(SymbolSet t) {
		try{
			return (this.union(t)).getNumberOfElements()<(this.getNumberOfElements()+t.getNumberOfElements());
		} catch(Exception e) {
			e.printStackTrace();
		}
		return true;
	}
	
	/**
	 * 
	 * @param str an arbitrary Symbol
	 * @return true if the str is contained in this SymbolSet
	 */
	protected boolean contains(String str) {
		return ts.contains(str);
	}
	
	protected static TreeSet<SymbolSet> createSymbolSets(DiscreteAlphabet alph) throws Exception {
		TreeSet<SymbolSet> symbolset;
		SymbolSet[] symbolsetarr = new SymbolSet[(int)(Math.pow(2,alph.length()))-1];
		int counter = 0;
		for(int i=0;i<alph.length();i++) {
			int countAtStartOfIteration = counter;
			for(int j=0;j<countAtStartOfIteration;j++) {
				symbolsetarr[counter] = symbolsetarr[j].union(new SymbolSet(alph, alph.getSymbolAt(i)));
				counter++;
			}
			symbolsetarr[counter]=new SymbolSet(alph, alph.getSymbolAt(i));
			counter++;
		}
		symbolset = new TreeSet<SymbolSet>();
		for(int i=0;i<symbolsetarr.length;i++)
			symbolset.add(symbolsetarr[i]);
		return symbolset;
	}
	
	public TreeSet<SymbolSet> createSymbolSubSets() throws Exception {
		String[] d = new String [1];
		d = ts.toArray(d);
		TreeSet<SymbolSet> symbolset;
		SymbolSet[] symbolsetarr = new SymbolSet[(int)(Math.pow(2,ts.size()))-1];
		int counter = 0;
		for(int i=0;i<ts.size();i++) {
			int countAtStartOfIteration = counter;
			for(int j=0;j<countAtStartOfIteration;j++) {
				symbolsetarr[counter] = symbolsetarr[j].union(new SymbolSet(alph, d[i]));
				counter++;
			}
			symbolsetarr[counter]=new SymbolSet(alph, d[i]);
			counter++;
		}
		symbolset = new TreeSet<SymbolSet>();
		for(int i=0;i<symbolsetarr.length;i++)
			symbolset.add(symbolsetarr[i]);
		return symbolset;
	}
	
	public TreeSet<SymbolSet> createSymbolMinimalSubSets() throws Exception {
		String[] d = new String [1];
		d = ts.toArray(d);
		TreeSet<SymbolSet> symbolset;
		SymbolSet[] symbolsetarr = new SymbolSet[(int)(Math.pow(2,ts.size()))-1];
		int counter = 0;
		for(int i=0;i<ts.size();i++) {
			int countAtStartOfIteration = counter;
			for(int j=0;j<countAtStartOfIteration;j++) {
				symbolsetarr[counter] = symbolsetarr[j].union(new SymbolSet(alph, d[i]));
				counter++;
			}
			symbolsetarr[counter]=new SymbolSet(alph, d[i]);
			counter++;
		}
		symbolset = new TreeSet<SymbolSet>();
		for(int i=0;i<symbolsetarr.length;i++) {
			if(symbolsetarr[i].getElements()[0].equals(this.getElements()[0]))
				symbolset.add(symbolsetarr[i]);
		}
		return symbolset;
	}
	
	public TreeSet<String> createSymbolMinimalSubSetsAsStrings() throws Exception {
		String[] d = new String [1];
		d = ts.toArray(d);
		TreeSet<String> symbolset;
		SymbolSet[] symbolsetarr = new SymbolSet[(int)(Math.pow(2,ts.size()))-1];
		int counter = 0;
		for(int i=0;i<ts.size();i++) {
			int countAtStartOfIteration = counter;
			for(int j=0;j<countAtStartOfIteration;j++) {
				symbolsetarr[counter] = symbolsetarr[j].union(new SymbolSet(alph, d[i]));
				counter++;
			}
			symbolsetarr[counter]=new SymbolSet(alph, d[i]);
			counter++;
		}
		symbolset = new TreeSet<String>();
		for(int i=0;i<symbolsetarr.length;i++) {
			if(symbolsetarr[i].getElements()[0].equals(this.getElements()[0]))
				symbolset.add(symbolsetarr[i].toSparseString());
		}
		return symbolset;
	}
	
	
	public double getScore() {
		return score;
	}
	
	public void setScore(double score) {
		this.score=score;
	}
	
	public String toSparseString() {
		String res = "";
		Iterator<String> it = ts.iterator();
		while( it.hasNext() ){
			res+=it.next();
		}
		return res;
	}
	
	public int size() {
		return ts.size();
	}
	
}
