package de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.inhomogeneous;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.NumberFormat;


import de.jstacs.InstantiableFromParameterSet;
import de.jstacs.NotTrainedException;
import de.jstacs.data.DataSet;
import de.jstacs.data.WrongLengthException;
import de.jstacs.data.sequences.Sequence;
import de.jstacs.io.NonParsableException;
import de.jstacs.io.XMLParser;
import de.jstacs.parameters.InstanceParameterSet;
import de.jstacs.results.NumericalResult;
import de.jstacs.results.NumericalResultSet;
import de.jstacs.results.Result;
import de.jstacs.results.SimpleResult;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.AbstractParsimoniousModel;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures.ParsimoniousElement;
import de.jstacs.sequenceScores.statisticalModels.trainable.variableStructure.parsimonious.dataStructures.ParsimoniousElement.LearningMethod;
import de.jstacs.utils.Normalisation;

/**
 * This class represents Inhomogeneous Parsimonious Markov Models (PMMs).
 * @author Ralf Eggeling<eggeling@informatik.uni-halle.de>
 *
 */
public class InhomogeneousPMM extends AbstractParsimoniousModel
		implements InstantiableFromParameterSet {
	
	protected static final String XML_TAG = "InhomogeneousParsimoniousMarkovModel";

	/**
	 * the parameter set of the full model
	 */
	private InhPMMParameterSet params;
	
	
	/**
	 * Constructs an inhomogeneous parsimonious Markov model (PMM) from a 
	 * @param params given parameter set.
	 * @throws Exception
	 */
	public InhomogeneousPMM(InhPMMParameterSet params) throws Exception {
		super(params.getAlphabetContainer(),params.getLength());
		try{
			this.params = params.clone();
		} catch(CloneNotSupportedException cnse) {
			System.out.println("ParameterSet could not be cloned. Use reference instead");
			this.params=params;
		}
		tree = new ParsimoniousElement[params.getLength()];
		byte order = (Byte)params.getParameterAt(0).getValue();
		for(int i=0;i<tree.length;i++) {
			tree[i] = new ParsimoniousElement(params.getAlphabetContainer(),(byte)Math.min(i,order),(Double)params.getParameterAt(1).getValue(),(Double)params.getParameterAt(2).getValue());
			tree[i].setElementNumber(i);
		}
		reader = null;
		writer = null;
		learningMethod = ParsimoniousElement.LearningMethod.MAXIMIZATION;
//		countHistory = new StringBuffer();
	}
	
	
	/**
	 * The standard constructor for the interface {@link de.jstacs.Storable}.
	 * Creates a new {@link InhomogeneousPMM} out of a {@link StringBuffer}.
	 * 
	 * @param sb
	 *            the {@link StringBuffer} to be parsed
	 * 
	 * @throws NonParsableException
	 *             is thrown if the {@link StringBuffer} could not be parsed
	 */	
	public InhomogeneousPMM(StringBuffer sb) throws NonParsableException {
		super(XMLParser.extractForTag(sb,XML_TAG));
		//XMLParser.extractForTag(sb,XML_TAG)
	}

	protected void fromXML(StringBuffer xml) throws NonParsableException {
		this.params = XMLParser.extractObjectForTags(xml, "params", InhPMMParameterSet.class);
		this.tree = XMLParser.extractObjectForTags(xml,"parsimoniousElements",ParsimoniousElement[].class);
		this.trained = XMLParser.extractObjectForTags(xml, "trained", boolean.class);
		this.learningMethod = XMLParser.extractObjectForTags(xml, "learningMethod", LearningMethod.class);
		this.numberOfSamples = XMLParser.extractObjectForTags(xml, "numberOfSamples", Integer.class);
		this.sampleFilePath =  XMLParser.extractObjectForTags(xml, "sampleFilePath", String.class);
		alphabets = params.getAlphabetContainer();
		length = params.getLength();
		}

	public void drawParameters(DataSet data) throws  Exception {
		for(int i=0;i<tree.length;i++) {
			tree[i].sampleStructureAndParameters(data.getInfixDataSet(i-Math.min(i,(Byte)params.getParameterAt(0).getValue()),1+Math.min(i,(Byte)params.getParameterAt(0).getValue())), null);
		}
		this.trained=true;
	}
	
	public StringBuffer toXML() {
		StringBuffer xml = new StringBuffer();
		XMLParser.appendObjectWithTags(xml, params, "params");
		XMLParser.appendObjectWithTags(xml, tree, "parsimoniousElements");
		XMLParser.appendObjectWithTags(xml, trained, "trained");
		XMLParser.appendObjectWithTags(xml, learningMethod, "learningMethod");
		XMLParser.appendObjectWithTags(xml, numberOfSamples, "numberOfSamples");
		XMLParser.appendObjectWithTags(xml, sampleFilePath, "sampleFilePath");
		XMLParser.addTags(xml, XML_TAG);
		return xml;
	}

	public String getInstanceName() {
		return "Inhomogeneous parsimonious Markov model of order "+params.getParameterAt(0).getValue();
	}

	@Override
	public double getLogPriorTerm() throws Exception {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public NumericalResultSet getNumericalCharacteristics() throws Exception {
		return new NumericalResultSet(new NumericalResult("numberOfLeaves","The total number of leaves of the inh. PMM",this.getNumberOfLeaves()));
	}

	public double getLogProbFor(Sequence seq, int start, int end) throws Exception {
		if(end-start+1!=length)
			throw new WrongLengthException("Sequence and model length do not match: "+(end-start+1)+"!="+length);
		if(inSamplingMode||learningMethod==LearningMethod.MAXIMIZATION) {
			return this.getLogProbForWholeSeq(seq.getSubSequence(start, end-start+1));
		} else {
			reader = new BufferedReader( new FileReader( new File(sampleFilePath) ) ) ;
			double res[] = new double[numberOfSamples];
			for(int i=0;i<numberOfSamples;i++) {
				this.parse(reader.readLine());
				res[i]=this.getLogProbForWholeSeq(seq.getSubSequence(start, end-start+1));
			}
			reader.close();
			return Normalisation.logSumNormalisation(res)-Math.log(numberOfSamples);
		}
	}
	
	private double getLogProbForWholeSeq(Sequence seq) throws WrongLengthException, NotTrainedException {
		double res = 0;
		for(int i=0;i<seq.getLength();i++) {
			double p = tree[i].getLogProbFor(seq, i);
			res += p;
		}
		return res;
	}

	public boolean isTrained() {
		return true;
	}
	
	public String getXMLTag() {
		return XML_TAG;
	}
	
	public InhomogeneousPMM clone() throws CloneNotSupportedException {
		InhomogeneousPMM cl = (InhomogeneousPMM)super.clone();
		cl.params = params.clone();
		cl.tree = new ParsimoniousElement[tree.length];
		for(int i=0;i<tree.length;i++) {
			if(tree[i]!=null)
				cl.tree[i] = tree[i].clone();
		}
		cl.trained=trained;
		return cl; 
	}
	
	public int getNumberOfLeaves() {
		int res = 0;
		for(int i=0;i<tree.length;i++) {
			res+= tree[i].getNumberOfLeaves();
		}
		return res;
	}

	public void train(DataSet data, double[] weights) throws Exception {
		for(int i=0;i<tree.length;i++) {
				DataSet dataForThisTree = data.getInfixDataSet(i-Math.min(i,(Byte)params.getParameterAt(0).getValue()),1+Math.min(i,(Byte)params.getParameterAt(0).getValue()));
				tree[i].mt=usedModelType;
				tree[i].maximizeStructureAndParameters(dataForThisTree, weights, usedStructureScore, usedParameterEstimate);
			}
		trained = true;
	}
	
	public void train(DataSet data, double[] weights, ParsimoniousElement.StructureScore structureScore, ParsimoniousElement.ParameterEstimate parameterEstimate) throws Exception {
		usedStructureScore = structureScore;
		usedParameterEstimate = parameterEstimate;
		train(data,weights);
	}
	
	public void train(DataSet data, ParsimoniousElement.StructureScore structureScore, ParsimoniousElement.ParameterEstimate parameterEstimate) throws Exception {
		usedStructureScore = structureScore;
		usedParameterEstimate = parameterEstimate;
		train(data);
	}
	
	public byte getOrder() {
		return (Byte)params.getParameterAt(0).getValue();
	}
	
	public double getESS() {
		return (Double)params.getParameterForName("ESS").getValue();
	}

	public double getLogScoreOfCurrentStructure() {
		double res = 0;
		for(int i=0;i<tree.length;i++) {
			res += tree[i].getScore();
		}
		return res;
	}

	@Override
	public String toString(NumberFormat nf) {
		return null;
	}
	
	public InstanceParameterSet<? extends InstantiableFromParameterSet> getCurrentParameterSet() throws Exception {
		return params.clone();
	}
	
	public void setSampleFilePath(String s) {
		this.sampleFilePath=s;
	}

}
