package de.jstacs.utils.functions;
import de.jtem.numericalMethods.calculus.specialFunctions.Gamma;

/**
 * This class provides basic functionality for computing Bayes-scores. 
 * All methods are static and public, so they can be used in arbitrary contexts. 
 * @author Ralf Eggeling <eggeling@informatik.uni-halle.de>
 *
 */

public class Bayesian {
	
	/**
	 * 
	 * @param N the data vector
	 * @param alpha the vector of hyperparameters
	 * @return the Bayes Factor, i.e. \frac{Beta(N+\alpha)}{Beta(\alpha)}
	 */
	public static double bayesFactor(double[] N, double[] alpha) {
		return Math.exp(logOfBayesFactor(N,alpha));
	}
	
	/**
	 * 
	 * @param N the data vector
	 * @param alpha the vector of hyperparameters
	 * @return the logarithm of the Bayes Factor, i.e. \log(\frac{Beta(N+\alpha)}{Beta(\alpha)})
	 */
	public static double logOfBayesFactor(double[] N, double[] alpha) {
	//	System.out.println(logOfMultinomialBeta(N,alpha)+"\t"+logOfMultinomialBeta(alpha));
		return logOfMultinomialBeta(N,alpha)-logOfMultinomialBeta(alpha);
	}
		
	/**
	 *  
	 * @param x An arbitrary vector of non-negative doubles
	 * @return logarithm of the multinomial beta function value
	 */
	public static double logOfMultinomialBeta(double[] x) {
		double sumOfX = 0;
		for(int i=0;i<x.length;i++) {
			sumOfX += x[i];
		}
		double res = 0;
		for(int i=0;i<x.length;i++) {
			res += Gamma.logOfGamma(x[i]);
		}
		return res - Gamma.logOfGamma(sumOfX);
	}
	
	/**
	 * 
	 * @param x An arbitrary vector of non-negative doubles
	 * @return multinomial beta function value
	 */
	public static double multinomialBeta(double[] x) {
		return Math.exp(logOfMultinomialBeta(x));
	}
	
	/**
	 * 
	 * @param x An arbitrary vector of non-negative doubles
	 * @param y An arbitrary vector of non-negative doubles
	 * @return logarithm of the multinomial beta function value of the sum of x and y
	 */
	public static double logOfMultinomialBeta(double[] x, double[] y) {
		double[] z = new double[x.length];
		for(int i=0;i<x.length;i++) {
			z[i] = x[i]+y[i];
		}
		return logOfMultinomialBeta(z);
	}
	
	/**
	 * 
	 * @param x An arbitrary vector of non-negative doubles
	 * @param y An arbitrary vector of non-negative doubles
	 * @return logarithm of the multinomial beta function value of the sum of x and y	 
	 */
	public static double multinomialBeta(double[] x, double[] y) {
		return Math.exp(logOfMultinomialBeta(x,y));
	}
	

}
