package de.jstacs.utils.functions;
import de.jstacs.data.alphabets.DiscreteAlphabet;
import de.jtem.numericalMethods.calculus.specialFunctions.Gamma;

/**
 * Contains functions related to the Normalized Maximum Likelihood (NML).  
 * @author Ralf Eggeling <eggeling@informatik.uni-halle.de>
 * 
 */

public class NML {
	
	/**
	 * Computes the NML normalizing constant for the multinomial distribution (Kontkanen & Myllymaki 2007)
	 * @param alph the alphabet of the multinomial distribution
	 * @param sampleSize the size of the vector
	 * @return the multinomial normalizing constant table
	 */
	public static double[][] multinomialaNormalizingConstantTable(DiscreteAlphabet alph, int sampleSize) {
		double[][] res = new double[sampleSize+1][(int)alph.length()];
		init(res);
		for(int n=1;n<res.length;n++) {
			for(int r = 2;r<res[n].length;r++) {
				res[n][r] = Math.log(Math.exp(res[n][r-1]) + ((double)n)/(r-1)*Math.exp(res[n][r-2]));
			}
		}		
		for(int n=1000;n<res.length;n++) {
			for(int r=1;r<res[n].length;r++) {
				res[n][r] = SzpankowskiApproximation(r+1,n);
			}
		} 
		return res;
	}
	
	
	private static void init(double[][] res) {
		for(int r=0;r<res[0].length;r++) {
			res[0][r] = 0;
		}
		for(int n=0;n<res.length;n++) {
			res[n][0] = 0;
		}
		for(int n=1;n<res.length;n++) {		
			res[n][1] = 0;
			for(int m=0;m<=n;m++) {
				res[n][1] += Combinatorics.binomialCoefficient(n, m)*Math.pow((((double)m)/n),m)*Math.pow((((double)(n-m))/n),n-m);
			}
			res[n][1] = Math.log(res[n][1]);
		}
	}
	
	/**
	 * Computes the Szpankowski approximation of the multinomal NML
	 * @param K categories
	 * @param n number of data points
	 * @return multinomial NML constant by Szpankowski approximation for n data points and K categories
	 */
	public static  double SzpankowskiApproximation(int K, int n) {
		double res = 0;
		if(n==0)
			return res;
		res += (K-1)/2*Math.log(n/2);
		res += Math.log(Math.sqrt(Math.PI))-Gamma.logOfGamma(K/2);
		res += Math.sqrt(2)*K/(3*Math.sqrt(n))*Math.exp(Gamma.logOfGamma(K/2)-Gamma.logOfGamma(K/2-0.5));
		res += ((3+K*(K-2)*(2*K+1))/36 - Math.exp(2*Gamma.logOfGamma(K/2)+2*Math.log(K)-Math.log(9)-2*Gamma.logOfGamma(K/2-0.5)))/n;
		return res;
	}
	
	/**
	 * Used for fsNML parameter estimation
	 * @param x any real number 
	 * @return ((x+1)/x)^x if x!=0, 1 otherwise
	 */
	public static double eTerm(int x) {
		if(x==0) {
			return 1;
		} else {
			return Math.pow((x+1.0)/x,x);
		}
	}
	
}
